#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define	FONTW	8
#define	FONTH	32
#define	TRS80	0
#define	INVERSE	0

uint8_t buff[256*FONTH];
int scanrows = 0;
int chars = 128;
int shr = 0;
int fourcol = 0;
int fonth = FONTH;

const char *bin(int val, int digits)
{
	static char buff[64];
	char *dst;

	dst = buff + digits;
	*dst-- = '\0';
	while (dst >= buff) {
		*dst-- = '0' + (val & 1);
		val >>= 1;
	}
	return buff;
}

const char *dup(int val, int digits)
{
	static char buff[64];
	char *dst;

	dst = buff + 2 * digits;
	*dst-- = '\0';
	while (dst >= buff) {
		*dst-- = '0' + (val & 1);
		*dst-- = '1';
		val >>= 1;
	}
	return buff;
}

void flipbits(uint8_t *buff, int w)
{
	switch (w) {
	case 6:
		*buff =
			((*buff >> 5) & 0x01) |
			((*buff >> 3) & 0x02) |
			((*buff >> 1) & 0x04) |
			((*buff << 1) & 0x08) |
			((*buff << 3) & 0x10) |
			((*buff << 5) & 0x20);
		break;
	case 7:
		*buff =
			((*buff >> 6) & 0x01) |
			((*buff >> 4) & 0x02) |
			((*buff >> 2) & 0x04) |
			((*buff << 2) & 0x10) |
			((*buff << 4) & 0x20) |
			((*buff << 6) & 0x40);
		break;
	case 8:
		*buff =
			((*buff >> 7) & 0x01) |
			((*buff >> 5) & 0x02) |
			((*buff >> 3) & 0x04) |
			((*buff >> 1) & 0x08) |
			((*buff << 1) & 0x10) |
			((*buff << 3) & 0x20) |
			((*buff << 5) & 0x40) |
			((*buff << 7) & 0x80);
		break;
	}
}

int main(int argc, char **argv)
{
	char *filename = NULL;
	FILE *fp;
	int bytes_read;
	int n, x, y, z;

	if (argc < 2) {
		printf("Usage: bin [options] filename\n");
		return 1;
	}
	for (n = 1; n < argc; n++) {
		if (!strcmp(argv[n], "-4")) {
			fourcol = 1;
			continue;
		}
		if (!strcmp(argv[n], "-s")) {
			++n;
			shr = atoi(argv[n]);
			continue;
		}
		if (!strcmp(argv[n], "-r")) {
			scanrows = 1;
			continue;
		}
		if (!strcmp(argv[n], "-f")) {
			chars = 256;
			continue;
		}
		if (!strcmp(argv[n], "-h")) {
			++n;
			fonth = atoi(argv[n]);
			continue;
		}
		filename = argv[n];
	}
	if (!filename) {
		fprintf(stderr, "No filename found\n");
		return 1;
	}
	fp = fopen(filename, "rb");
	if (NULL == fp) {
		fprintf(stderr, "Cannot open '%s'\n", argv[1]);
		return 1;
	}
	z = 0;
	bytes_read = fread(buff, 1, 256*fonth, fp);
	if (bytes_read == 128*fonth)
		chars = 128;
	else if (bytes_read != 256*fonth)
		return 1;

#if	TRS80
	for (z = 127; z >= 0; z--) {
		if (z == ',' || z == ';' || z == 'g' || z == 'j' || z == 'p' || z == 'q' || z == 'y') {
			memset(buff + z * fonth + 0, 0, 2);
			memmove(buff + z * fonth + 2, buff + z * 8, 8);
			memset(buff + z * fonth + 10, 0, fonth - 10);
		} else {
			memmove(buff + z * fonth - 1, buff + z * 8, 8);
			memset(buff + z * fonth + 7, 0, fonth - 7);
		}
	}
#endif

	if (scanrows) {
		for (y = 0; y < fonth; y++) {
			printf("\n\t\t\t\t' scanrow %d\n", y);
			for (z = 0; z < chars; z++) {
				n = z * fonth + y;
				if (z < 128) {
					buff[n] = buff[n] >> shr;
					flipbits(buff + n, FONTW);
				} else {
#if	TRS80
					if (y < fonth*1/3)
						buff[n] = (z & 0x01 ? 0x07 : 0) | (z & 0x02 ? 0x38 : 0);
					else if (y < fonth*2/3)
						buff[n] = (z & 0x04 ? 0x07 : 0) | (z & 0x08 ? 0x38 : 0);
					else
						buff[n] = (z & 0x10 ? 0x07 : 0) | (z & 0x20 ? 0x38 : 0);
#else
#if	INVERSE
					buff[n] = buff[n - 128*fonth] ^ ((1 << FONTW) - 1);
#else
					buff[n] = buff[n] >> shr;
					flipbits(buff + n, FONTW);
#endif
#endif
				}
				if (fourcol)
					printf("\t\tword\t%%%s\t' ", dup(buff[n], FONTW));
				else
					printf("\t\tbyte\t%%%s\t' ", bin(buff[n], FONTW));
				for (x = 0; x < FONTW; x++) {
					printf("%c", (buff[n] & (1 << x)) ? '#' : '.');
				}
				printf("\n");
			}
	/*		printf("\t\tbyte\t0\t' filler\n"); */
		}
	} else {
		for (z = 0; z < chars; z++) {
			printf("\n\t\t\t\t' chr$(%02x)\n", z);
			for (y = 0; y < fonth; y++) {
				n = z * fonth + y;
				if (z < 128) {
					buff[n] = buff[n] >> shr;
					flipbits(buff + n, FONTW);
				} else {
#if	TRS80
					if (y < fonth*1/3)
						buff[n] = (z & 0x01 ? 0x07 : 0) | (z & 0x02 ? 0x38 : 0);
					else if (y < fonth*2/3)
						buff[n] = (z & 0x04 ? 0x07 : 0) | (z & 0x08 ? 0x38 : 0);
					else
						buff[n] = (z & 0x10 ? 0x07 : 0) | (z & 0x20 ? 0x38 : 0);
#else
#if	INVERSE
					buff[n] = buff[n - 128*fonth] ^ ((1 << FONTW) - 1);
#else
					buff[n] = buff[n] >> shr;
					flipbits(buff + n, FONTW);
#endif
#endif
				}
				if (fourcol)
					printf("\t\tword\t%%%s\t' ", dup(buff[n], FONTW));
				else
					printf("\t\tbyte\t%%%s\t' ", bin(buff[n], FONTW));
				for (x = 0; x < FONTW; x++) {
					printf("%c", (buff[n] & (1 << x)) ? '#' : '.');
				}
				printf("\n");
			}
	/*		printf("\t\tbyte\t0\t' filler\n"); */
		}
	}
	fclose(fp);
}
